<!DOCTYPE kpartgui>
<kpartgui name="ksysguard" version="5">
<MenuBar>
  <Menu name="file"><text>&amp;File</text>
    <Action name="new_worksheet"/>
    <Action name="import_worksheet"/>
    <Action name="get_new_worksheet"/>
    <Action name="upload_worksheet"/>
    <Action name="export_worksheet"/>
    <Action name="remove_worksheet"/>
    <Separator/>
    <Action name="connect_host"/>
  </Menu>
  <Menu name="view"><text>&amp;View</text>
    <Action name="configure_sheet"/>
  </Menu>
  <Menu name="help"><Action name="switch_application_language"/></Menu>
  <Menu name="settings"><text>&amp;Settings</text>
  </Menu>
</MenuBar>
<ToolBar name="mainToolBar" noMerge="1" hidden="true"><text>Main Toolbar</text>
    <Action name="new_worksheet"/>
    <Action name="import_worksheet"/>
    <Action name="get_new_worksheet"/>
    <Action name="export_worksheet"/>
    <Separator/>
    <Action name="connect_host"/>
    <Separator/>
    <Action name="configure_sheet"/>
</ToolBar>
</kpartgui>

